"""Indicators module."""

# The actual code for importing virtual modules is in the top-level __init__.
# This is for import reasons: we need to make sure all normal indicators are created before.
from __future__ import annotations
