"""
====================
Data files for xclim
====================

JSON and YAML definitions for virtual modules and internationalisation support.

Currently, the following virtual modules are defined:
  * ANUCLIM (`anuclim.yml`)
  * CF (`cf.yml`)
  * ICCLIM (`icclim.yml`)

And the following languages are supported:
  * English (default)
  * French (`fr.json`)

These files are used by xclim to define new indicators and to provide translations for the user interface.

Additionally, this package contains the following data files:
  * `schema.yml`: YAML schema for detailing class definitions used for indicators.
  * `variables.yml`: YAML schema defining the variables and their metadata used in the indicator definitions.
"""
